/*
 *   FITSCCD:   read fits files - and near-fits files
 *
 *	ToDO:
 *		allow bscale/bzero/crpix etc. to be overriden
 *		even if fits file; i.e. make their
 *		defaults blanks, and use hasvalue() to detect mode.
 *
 *	 7-sep-89       started coding on 3b1
 *	28-jan-90	[32] V1.0 finished finally
 *       8-mar-90       V1.1 fixed bug on suns - added slice=
 *	16-mar-90       V2.0 New fits I/O routines, options=	PJT
 *	 1-oct-90	V3.0 Also allow Sault's fitsio routines PJT
 *	11-oct-90	V3.1 allow near-fits files PJT
 *       2-may-91       V3.2 fixed Xmin() bug       PJT
 *	10-sep-91       V3.3 allow blank value re-substitution	PJT
 *                       3.3a   also check datamin/max          PJT
 *	 7-mar-92	 3.4  gcc2.0 happy - new style		PJT
 *	25-aug-92	 3.5 fixed minmax bug if not in fits file PJT
 *      31-mar-93           c  fixed bug in fitrdhda() calling  PJT
 *	 9-sep-93           d  fixed bug - forgot bytepix in rawopen?
 *      29-jun-94           e  new fts_rdata() arg order (old code)   PJT
 *	26-jul-94	    g  allow bitpix=8 in rawopen	      PJT
 *	11-oct-95	    h  fixed bug in writing CTYPE3		pjt
 */

#include <stdinc.h>
#include <getparam.h>
#include <image.h>
#include <fitsio.h>

string defv[] = {
    "in=???\n		Input fits file",
    "out=???\n		Output image file",
    "planes=\n          Planes to select from fits cube [all]", 
    "blocking=1\n	Extra blocking factor for input (blocksize/2880)",
    "mode=fits\n        Format mode of input file {fits,raw}",
    "blocksize=1\n      Blocksize in raw mode",
    "offset=0\n         Data offset in raw mode",
    "bitpix=32\n        Bitpix in raw mode {16,32,-32}",
    "naxis=\n           Length of the axes in raw mode",
    "cdelt=\n           Pixel separation in raw mode (2/naxis)",
    "crpix=0,0,0\n      Reference pixel in raw mode (0)",
    "crval=0,0,0\n      Coordinates of reference pixel in raw mode (0)",
    "bscale=1\n         Scale conversion factor in raw mode (1)",
    "bzero=0\n          Offset conversion factor in raw mode (0)",
    "blank=\n           Blank value re-substitution value?",
    "VERSION=3.5h\n	11-oct-95 PJT",
    NULL,
};

string usage = "convert (near)fits files into images";

#define MAXPLANES 512

void make_fitheader(), make_rawheader();
FITS *rawopen();

void nemo_main()
{
    stream outstr;
    FITS *fitsfile;
    int ndim=3, naxis[3], nx, ny, nz, i, j, k, npl, p, planes[MAXPLANES];
    int nbval=0;
    real bval_out, rmin, rmax, tmp;
    FLOAT *buffer, *bp, bval_in;  /* fitsio- is in FLOAT !!! */
    imageptr iptr;
    string mode, tmps;
    bool   Qblank;

    outstr = stropen(getparam("out"),"w");   /* open image file for output */
    npl = nemoinpi(getparam("planes"),planes,MAXPLANES);
    mode = getparam("mode");
    tmps = getparam("blank");
    Qblank = (*tmps != NULL);
    if (Qblank) {    
        bval_in = (FLOAT) getdparam("blank");
        bval_out = 0.0;
    }
    
    if (streq(mode,"fits"))
        fitsfile = fitopen(getparam("in"),"old",ndim,naxis);
    else if (streq(mode,"raw"))
        fitsfile = rawopen(getparam("in"),"old",ndim,naxis);
    else
        error("Illegal mode %s: must be one of {fits,raw}",mode);

    if (fitsfile==NULL) error("Could not open file %s in %s mode\n",
               getparam("in"),mode);
    nx = naxis[0];
    ny = naxis[1];
    nz = (npl>0) ? npl : naxis[2];
    if (nx*ny*nz==0) error("Bad fits image: nx*ny*nz=0");
    if (nx==1) error("Fits image: nx=1 ???");
    if (ny==1) error("Fits image: ny=1 ???");
    if (nz > 1) {
        dprintf(0,"FITS file: Image size %d %d %d\n",nx,ny,nz);
        create_cube(&iptr,nx,ny,nz);
    } else {
        dprintf(0,"FITS file: Image size %d %d\n",nx,ny);
        create_image(&iptr,nx,ny);
    }
    if (iptr==NULL) error("No memory to allocate image");

    if (streq(mode,"fits"))
        make_fitheader(fitsfile,iptr);
    else if (streq(mode,"raw"))
        make_rawheader(fitsfile,iptr);
    else
        error("Never Reached");

    buffer = (FLOAT *) allocate(nx*sizeof(FLOAT));

    rmin = HUGE;
    rmax = -HUGE;
    for (k=0; k<nz; k++) {          /* loop over all/selected planes */
        p = (npl>0) ? planes[k] : k;        /* select plane number */
        dprintf(1,"Reading plane %d\n",p);
        fitsetpl(fitsfile,1,&p);
        for (j=0; j<ny; j++) {      /* loop over all rows */
            fitread(fitsfile,j,buffer);     /* read it from fits file */
            for (i=0, bp=buffer; i<nx; i++, bp++) {   /* stuff it in memory */
                if (Qblank) {
                    if (*bp == bval_in) {
                        nbval++;
                        *bp = bval_out;
                    }
                }
                tmp = CubeValue(iptr,i,j,k) = *bp;
                rmin=MIN(rmin,tmp);
                rmax=MAX(rmax,tmp);
            }
        }
    }
    if (rmin != MapMin(iptr)) {
        warning("Setting map minimum to %g",rmin);
        MapMin(iptr) = rmin;
    }
    if (rmax != MapMax(iptr)) {
        warning("Setting map maximum to %g",rmax);
        MapMax(iptr) = rmax;
    }
    fitclose(fitsfile);
    write_image(outstr,iptr);
    if (Qblank) {
        if (nbval==0)
            printf("There were no blank values in the image\n");
        else
            printf("There were %d blank values in the image\n",nbval);
    }
}

FITS *rawopen(name,status,naxis,nsize)
string name, status;
int naxis, nsize[];
{
    FITS *f;
    int i,ndim,bitpix;
    
    f = (FITS *) allocate(sizeof(FITS));

    if (streq(status,"old")){       /*  handle an old file */
        f->fd = stropen(name,"r");
        ndim = nemoinpi(getparam("naxis"),nsize,naxis);
        if (ndim < 1) error("naxis= returns %d from parser",ndim);
        f->naxis = ndim;
        for (i=ndim; i<naxis; i++) nsize[i] = 1;
        for (i=0; i<ndim; i++)  f->axes[i] = nsize[i];
        for (i=ndim; i<MAXNAX; i++) f->axes[i] = 1;
        bitpix = getiparam("bitpix");
        fit_setbitpix(bitpix);
        if (bitpix==8)        f->type = TYPE_8INT;
        else if (bitpix==16)  f->type = TYPE_16INT;
        else if (bitpix==32)  f->type = TYPE_32INT;
        else if (bitpix==-32) f->type = TYPE_FLOAT;
        else error("Invalid bitpix: %d",bitpix);
        f->bytepix = ABS(bitpix)/8;
        f->status = STATUS_OLD;
        f->bscale = getdparam("bscale");
        f->bzero  = getdparam("bzero");
        f->offset = getiparam("offset");
        f->ncards = 0;
    } else
        error("Cannot rawopen %s as \"%s\"\n",name,status);
    fit_setblocksize( getiparam("blocksize") * getiparam("blocking") );
    return f;
}

void make_rawheader(fitsfile,iptr)
FITS *fitsfile;
imageptr iptr;
{
    int ndim, n, tmpr[3], nemoinpd();
    
    ndim = (Nz(iptr) > 1) ? 3 : 2;

    n = nemoinpd(getparam("cdelt"),tmpr,3);
    if (n == 0) {
        Dx(iptr) = 2.0/Nx(iptr);
        Dy(iptr) = 2.0/Ny(iptr);
        Dz(iptr) = 2.0/Nz(iptr);
    } else if (n == 3) {
        Dx(iptr) = tmpr[0];
        Dy(iptr) = tmpr[1];
        Dz(iptr) = tmpr[2];
    } else
        error("cdelt= needs either 3 or 0 entries");

    Xmin(iptr) = 0.0;
    Ymin(iptr) = 0.0;
    Zmin(iptr) = 0.0;

}

void make_fitheader(fitsfile,iptr)
FITS *fitsfile;
imageptr iptr;
{
    int nz, tmpi;
    real crpix1, crpix2, crpix3;
    FLOAT tmpr;
    char ctype[32], *scopy();

    nz = Nz(iptr);

    fitrdhdr(fitsfile,"CRVAL1",&tmpr,0.0); Xmin(iptr) = tmpr;
    fitrdhdr(fitsfile,"CRVAL2",&tmpr,0.0); Ymin(iptr) = tmpr;
    if (nz>1) fitrdhdr(fitsfile,"CRVAL3",&tmpr,0.0); Zmin(iptr) = tmpr;

    fitrdhdr(fitsfile,"CDELT1",&tmpr,1.0); Dx(iptr)=tmpr;
    fitrdhdr(fitsfile,"CDELT2",&tmpr,1.0); Dy(iptr)=tmpr;
    if (nz>1) fitrdhdr(fitsfile,"CDELT3",&tmpr,1.0); Dz(iptr)=tmpr;

    fitrdhdr(fitsfile,"CRPIX1",&tmpr,1.0); crpix1 = tmpr;
    fitrdhdr(fitsfile,"CRPIX2",&tmpr,1.0); crpix2 = tmpr;
    if (nz>1) fitrdhdr(fitsfile,"CRPIX3",&tmpr,1.0); crpix3 = tmpr;

    if (crpix1 != 1.0)
        Xmin(iptr) -= (crpix1-1.0)*Dx(iptr);
    if (crpix2 != 1.0)
        Ymin(iptr) -= (crpix2-1.0)*Dy(iptr);
    if (nz>1 && crpix3 != 1.0)
        Zmin(iptr) -= (crpix3-1.0)*Dz(iptr);

    fitrdhda(fitsfile,"CTYPE1",ctype,"");
    Namex(iptr) = scopy(ctype);

    fitrdhda(fitsfile,"CTYPE2",ctype,"");
    Namey(iptr) = scopy(ctype);

    if (nz>1) {
           fitrdhda(fitsfile,"CTYPE3",ctype,"");
	   Namez(iptr) = scopy(ctype);
    }

    fitrdhdr(fitsfile,"DATAMIN",&tmpr,0.0); MapMin(iptr) = tmpr;
    fitrdhdr(fitsfile,"DATAMAX",&tmpr,0.0); MapMax(iptr) = tmpr;
    if (fitexhd(fitsfile,"BLANK")) {
        dprintf(0,"BLANK keyword exists\n");
        fitrdhdi(fitsfile,"BITPIX",&tmpi,-1);
        if (tmpi<0) {
            warning("FITS keyword BLANK not interpreted - BITPIX<0");
        } else {
            fitrdhdi(fitsfile,"BLANK",&tmpi,0);
            dprintf(0,"BLANK = %d\n",tmpi);
        }
    }
}

